"""Scarf instance input/output."""

import numpy as np
import json
import pickle
from scipy import io as sio

import scarf.instance

__all__ = ["save_json", "load_json", "save_mat", "load_excel",
           "save_pickle", "load_pickle"]


def save_json(ins, filename):
  """Save ScarfInstance to json format.

  Args:
    ins: a `ScarfInstance` object.
    filename: output file name.
  """
  with open(filename, mode="w") as g:
    hospital_pref_list = ins.hospital_pref_list.hospital_pref_list
    if ins.hospital_pref_list.is_ihp: 
      hospital_pref_list = [hospital_pref_list]
    json.dump(
        {
            "single_pref_list": ins.single_pref_list,
            "couple_pref_list": ins.couple_pref_list,
            "hospital_pref_list": hospital_pref_list,
            "hospital_cap": ins.hospital_cap
        }, g, indent=4
    )


def _element_check(obj):
  if isinstance(obj, list):
    return tuple(obj)
  elif isinstance(obj, int):
    return obj
  else:
    raise TypeError("Wrong!")


def load_json(filename):
  """Read instance from a json file of preferences.

  Args:
    filename: input json file name.
  Returns:
    A `ScarfInstance` object.
  """
  with open(filename) as f:
    all_fields = json.load(f)
  single_pref_list = all_fields["single_pref_list"]
  couple_pref_list = [[_element_check(hp) for hp in li] 
                      for li in all_fields["couple_pref_list"]]
  hospital_pref_list = [[_element_check(i) for i in li]
                        for li in all_fields["hospital_pref_list"]]
  if len(hospital_pref_list) == 1:
    hospital_pref_list = hospital_pref_list[0]
  hospital_cap = all_fields["hospital_cap"]
  return scarf.instance.ScarfInstance(
      single_pref_list=single_pref_list,
      couple_pref_list=couple_pref_list,
      hospital_pref_list=hospital_pref_list,
      hospital_cap=hospital_cap
  )


def save_mat(ins, filename):
  """Save part of the instance to MATLAB style .mat file.

  Save the constraint matrix, utility matrix, and right hand side vector into a
  '.mat' file for use in MATLAB.

  Args:
    ins: A `ScarfInstance`.
    filename: output filename with or without '.mat' extension.
  """
  sio.savemat(
        filename,
        {
            "A": ins.A.astype(np.int32),
            "U": ins.U,
            "b": np.array(ins.full_b()).reshape(-1, 1)
        }
  )


def load_excel(filename):
  """Obtain matching instance from excel."""
  raise NotImplementedError("Developer has been lazy.")


def save_pickle(ins, filename):
  """Save ScarfInstance to python's pickle format.

  Args:
    ins: a `ScarfInstance`.
    filename: output file name.
  """
  with open(filename, "wb") as g:
    pickle.dump(
        {
            "single_pref_list": ins.single_pref_list,
            "couple_pref_list": ins.couple_pref_list,
            "hospital_pref_list": ins.hospital_pref_list.hospital_pref_list,
            "hospital_cap": ins.hospital_cap
        }, g
    )


def load_pickle(filename):
  """Read instance from a python pickle file of preferences.

  Warning: As official python3 documentation has suggested, pickle format is NOT
  secure against adversarial attack. Please make sure you trust the source of the
  data file.

  Args:
    filename: json file name.
  Returns:
    A `ScarfInstance` object.
  """
  with open(filename, "rb") as f:
    all_data = pickle.load(f)
  return scarf.instance.ScarfInstance(
      single_pref_list=all_data["single_pref_list"],
      couple_pref_list=all_data["couple_pref_list"],
      hospital_pref_list=all_data["hospital_pref_list"],
      hospital_cap=all_data["hospital_cap"]
  )


"""
Scarf
=============================================
Solving stable matching problem with couples.

Example:

Suppose that we would like to solve a stable matching problem with 2 single
doctors (s0, s1), 2 couples (c0, c1), and 2 hospitals (h0, h1). We now create
preference lists of all doctors and hospitals, such that:
- 
- for a couple

Say, single doctor s0 ranks hospital h0 > h1 > h2, and single doctor s1 ranks
hospital h1 > h0 > unemployment > h2 (yep, she would rather be unemployed than
going to hospital h2.):

The input preferences list is a list of acceptable options, i.e. 
options that are better than unemployment, from the most preferrable to the least 
preferrable
---------------------------------------------
  >>> single_pref = [[0, 1, 2], 
  ...                [1, 0]]
---------------------------------------------
Different from single doctors, couples makes decision together. Couple c0 (consist of
members c00 and c01) ranks the plan "both go to h0" > "both go to h1" > 
"c00 to h1, c01 to h0" > "c00 to h0, c01 to h1" > "both to h2" > "both unemployed".
Similarly, c1 gives a list, which reads "both to h1" > "both to h2" > 
"c10 to h1, c11 unemplyed" > "both unemployed" > all other plans.

Use -1 to indicate unemployment option. The input preferences list is a list of acceptable
options, i.e. options that are better than "both unemployed", from the most preferrable to 
the least preferrable
---------------------------------------------
  >>> couple_pref = [[(0, 0), (1, 1), (1, 0), (0, 1), (2, 2)],
  ...                [(1, 1), (2, 2), (1, -1)]]
----------------------------------------------
Hospital also has a preference on individual doctors. Here we assume that all three
hospitals use the same preference order: s0 > c01 > s1 > c10 > c00 > c11 > vacancy.
The input preferences list is a list of acceptable doctors, i.e. options that are better than
leaving a seat vacant, from the most preferrable to the least preferrable. Single doctors
are represented by integers. The first member of a couple c is represented with a tuple (c, 0).
The second member of a couple c is represented with a tuple (c, 1).
----------------------------------------------
  >>> hospital_pref = [0, (0, 1), 1, (1, 0), (0, 0), (1, 1)]
----------------------------------------------
Alternatively, if hospitals have different preference over doctors, then we specify
a list of lists satisfying the above format
----------------------------------------------
  >>> hospital_pref = [list0, list1, list2]
----------------------------------------------
Now we specify the capacity of the three hospitals:
----------------------------------------------
  >>> hospital_cap = [2, 3, 1]
----------------------------------------------
Construct the stable matching instance and solve for a potentially fractional stable matching
----------------------------------------------
  >>> import scarf
  >>> S = scarf.create_instance(single_pref, couple_pref, hospital_pref, hospital_cap)
  >>> sol = scarf.solve(S)
----------------------------------------------
In the above example, the above procedure will find an integral stable matching.
However, for many large instances, especially the ones with many couples, Scarf's algorithm
will return a fractional solution. We can then apply Nguyen and Vohra's iterative rounding
algorithm to find an integral matching that may slightly violate the capacity constraint of
the hospitals:
----------------------------------------------
  >>> intsol = scarf.round(sol, S)
----------------------------------------------
Both scarf.solve and scarf.round returns an instance of scarf.ScarfSolution object. 
Please refer to the docstring of scarf.ScarfSolution to see different ways to access the solution.
"""

__author__ = "Dengwang Tang"

from scarf.instance import *
from scarf.io import *
from scarf.random import *


"""Scarf Algotihm util functions."""

import numpy as np


def check_single_preflist(s, pair_list):
  assert(np.all([len(p) == 2 and p[0] == s for p in pair_list]))
  assert(pair_list[-1] == (s, -1))


def check_couple_preflist(c, pair_list):
  assert(np.all([len(p) == 3 and p[0] == c for p in pair_list]))
  assert(pair_list[-1] == (c, -1, -1))


def check_hospital_preflist(h, pair_list):
  assert(np.all(h in p for p in pair_list))
  assert(pair_list[-1] == (-1, h))


def recover_pref_lists(num_single, num_couple, num_hospital, U,
                       pair_list):
  """Recover preference list from utility matrix."""
  assert(U.shape[0] == num_single + num_couple + num_hospital)
  num_hospital_pair = (num_hospital + 1) ** 2 - 1
  assert(U.shape[1] == len(pair_list))
  # Sort decend by negating the array
  orders_U = np.argsort(-U)
  single_pref_list, couple_pref_list, hospital_pref_list = [], [], []
  for s in range(num_single):
    cols = orders_U[s]
    single_s_pair_list = [pair_list[col] for col in cols if U[s][col] < 0]
    check_single_preflist(s, single_s_pair_list)
    # The last one is (s, num_hospital), remove
    single_pref_list.append([p[1] for p in single_s_pair_list[:-1]])
  for c in range(num_couple):
    cols = orders_U[num_single + c]
    couple_c_pair_list = [pair_list[col] for col in cols 
                          if U[num_single + c][col] < 0]
    check_couple_preflist(c, couple_c_pair_list)
    # The last one is (c, num_hospital, num_hospital), remove
    couple_pref_list.append([p[1:] for p in couple_c_pair_list[:-1]])
  for h in range(num_hospital):
    cols = orders_U[num_single + num_couple + h]
    hospital_h_pref_list = [pair_list[col] for col in cols
                            if U[num_single + num_couple + h][col] < 0]
    check_hospital_preflist(h, hospital_h_pref_list)
    # The last one is (s, num_hospital), remove
    hospital_pref_list.append(hospital_h_pref_list[:-1])
  return single_pref_list, couple_pref_list, hospital_pref_list


def create_hospital_pref_on_pairs(num_hospital, h, one_hospital_pref_list, 
                                  single_pref_list, couple_pref_list):
  """Create hospital's preference on pairs given preference on individuals."""
  pair_pref_list = []
  for i in one_hospital_pref_list:
    if isinstance(i, int):
      if h in single_pref_list[i]:
        pair_pref_list += [(i, h)]
    else:
      c, j = i  # couple c, member j
      # find out if this member is the better one
      member_j_position = one_hospital_pref_list.index((c, j))
      other_member_position = one_hospital_pref_list.index((c, 1 - j))
      is_better_member = member_j_position < other_member_position
      # filter out the pairs related to hospital h and (c, j)
      # where (c, j) is the worst member assigned to couple h in this pair
      def is_relavent_pair(p):
        return p[j] == h and not (p == (h, h) and is_better_member)
      couple_c_pref_list = list(filter(
          is_relavent_pair,
          couple_pref_list[c]
      ))
      pair_pref_list += [(c,) + p for p in couple_c_pref_list]

  return pair_pref_list


def check_stable(U, basis):
  """Check if a basis is ordinal basis for a utility matrix.

  Args:
    U: (m, n) utility matrix.
    basis: a list of m column indices

  Returns:
    True if `basis` is an ordinal basis of `U`
  """
  U_pt = U + np.tile(
    np.linspace(start=0.5, stop=0.0, num=U.shape[1]), 
    (U.shape[0], 1))
  rowmins = np.min(U[:, basis], axis=1, keepdims=True)
  U_dom = U <= rowmins
  return np.all(np.any(U_dom, axis=0))


def check_feasible(A, basis, alloc, b, tol=1e-6):
  """Check if a basis is a feasible basis for a polytope.

  Args:
    A: (m, n) constraint matrix.
    basis: a list of m column indices
    alloc: allocation vector
    b: the right hand side vector of size (m,)

  Returns:
    True if `basis` is a feasible basis of the polytope `Ax=b, x>=0`.
  """
  return np.all(np.dot(A[:, basis], alloc) <= np.array(b) + tol)


"""Random Instance Generators"""

import numpy as np

import scarf.instance

__all__ = ["gen_random_instance"]


def _id_2_tuple(num_single, idx):
  if idx < num_single:
    return idx
  else:
    couple_idx = idx - num_single
    return (couple_idx // 2, couple_idx % 2)


def gen_random_instance(num_single, num_couple, num_hospital,
                        num_additional_seat=0,
                        single_pref_len=0, couple_pref_len=0, ihp=True):
  """Generate a uniform random instance.

  Generate a stable matching instance where single doctor's preference lists are
  chosen uniformly at random, each hospital's ranking of doctors are also chosen
  at random. The couple's preference lists are chosen at random from all lists
  that are unemployment averse.

  Args:
    num_single: int
      Number of single doctors.
    num_couple: int
      Number of coupled doctors.
    num_hospital: int
      Number of hospitals.
    num_additional_seat: int, optional
      Number of total seats will be number of applicants plus num_additional_seat,
      provided that each hospital has at least one seat (Otherwise the hospital
      capacity is 1 for every hospital). Default is 0.
    single_pref_len: int, optional
      Length of single's preference list. Any hospital outside 
      preference list are worse than unemployment option. Default: generate full
      preference list.
    couple_pref_len: int, optional
      Length of couple's preference list. Any hospital pair outside 
      preference list are worse than unemployment option. Default: generate full
      preference list.
    ihp: bool, optional
      If True (default), hospitals will share the same preference on individual
      doctors. Otherwise hospitals will have independent preference lists.

  Returns:
    A `ScarfInstance` object.
  """
  num_applicant = num_single + 2 * num_couple
  single_pref_list = np.argsort(
      np.random.rand(num_single, num_hospital)
  ).tolist()
  if single_pref_len:
    for s in range(num_single):
      single_pref_list[s] = single_pref_list[s][:single_pref_len]
  num_hospital_pair = (num_hospital + 1) ** 2 - 1
  couple_pref_seed = np.random.rand(num_couple, num_hospital_pair)
  couple_pref_seed += np.array(
      [(pid // (num_hospital + 1) == num_hospital) or 
       (pid % (num_hospital + 1) == num_hospital)
       for pid in range(num_hospital_pair)], dtype=np.float64)
       # plans with unemployment is are ranked lower
  couple_pref_list = np.argsort(couple_pref_seed).tolist()
  wrap = lambda x: x if x < num_hospital else -1
  cpl = min(couple_pref_len,
            num_hospital_pair) if couple_pref_len > 0 else num_hospital_pair
  couple_pref_list = [
      [(wrap(pid // (num_hospital + 1)), 
      	wrap(pid % (num_hospital + 1))) for pid in li[:cpl]]
      for li in couple_pref_list
  ]
  if ihp:
    hospital_pref_list = np.argsort(
        np.random.rand(num_applicant)
    ).tolist()
    hospital_pref_list = [
        _id_2_tuple(num_single, i) for i in hospital_pref_list
    ]
  else:
    hospital_pref_list = np.argsort(
        np.random.rand(num_hospital, num_applicant)
    ).tolist()
    hospital_pref_list = [
        [_id_2_tuple(num_single, i) for i in li] for li in hospital_pref_list
    ]
  hosp_seat = np.random.randint(
      num_hospital,
      size=max(num_applicant - num_hospital + num_additional_seat, 0)
  )  # assign each seat randomly to a hospital
  hospital_cap = []
  for h in range(num_hospital):
    hospital_cap.append(int(sum(hosp_seat == h) + 1))
  
  return scarf.instance.ScarfInstance(
      single_pref_list=single_pref_list,
      couple_pref_list=couple_pref_list,
      hospital_pref_list=hospital_pref_list,
      hospital_cap=hospital_cap
  )




"""Scarf algorithm implementation"""

import numba as nb
import numpy as np
import scipy.optimize as spopt
from scipy.linalg import blas
import time

__all__ = ["scarf_solve", "iterative_rounding"]


@nb.njit('int32(int32[:])')
def my_argmin(vec):
  """Argmin function, return's largest index in the case of a tie."""
  return len(vec) - 1 - np.argmin(vec[-1:-1-len(vec):-1])


@nb.njit('void(boolean[:,:], int32[:,:], int64, int64)', parallel=True)
def _update_U_dom_row(U_dom, U, r, new_rowmin_col):
  """Update a row of the book-keeping variable U_dom."""
  new_rowmin = U[r, new_rowmin_col]
  if new_rowmin == 0:  # need to ensure left zeros LARGER
    U_dom[r, :new_rowmin_col] = U[r, :new_rowmin_col] < 0
    U_dom[r, new_rowmin_col:] = True  # the one's to the right are always smaller
  else:
    U_dom[r, :] = U[r, :] <= new_rowmin


@nb.njit('void(float64[:,:], float64[:], int64, int64, int64[:])', parallel=True)
def _update_A_b(A, b, row, pivot, oprows):
  """Subtract a multiple of the row A[row, :] from all other rows.

  Implemented with Numba's parallel loop.
  The rows that A[:, pivot] is pre-selected to reduce work.
  """
  for r0 in nb.prange(len(oprows)):
    r = oprows[r0]
    if r != row :
      b[r] -= A[r, pivot] * b[row]
      A[r, :] -= A[r, pivot] * A[row, :]


def _update_A_b_blas(A, b, row, pivot):
  """Subtract a multiple of the row A[row, :] from all other rows.

  Implemented with BLAS routines wrapped by scipy.
  """
  x = np.copy(A[:, pivot])
  x[row] = 0.0  # mask this row
  blas.daxpy(x=x, y=b, a=-b[row])  # y += a * x
  # A += alpha * x * A[row, :]; do it in transpose since A is in C-order.
  blas.dger(alpha=-1.0, x=A[row, :], y=x, a=A.T, overwrite_a=1)


def cardinal_pivot(A, b, card_basis, pivot):
  """Cardinal pivoting step.

  Introduce `pivot` into the cardinal basis `b` and remove a column j so that 
  `b` is still a cardinal basis.

  Args:
    A: an (m, n) matrix. Must be a non-negative numpy array.
  """
  candrows = np.nonzero(A[:, pivot] > 0)[0]
  if candrows.size == 0:
    raise RuntimeError("Cardinal pivot failure.")
  ratios = b[candrows] / A[candrows, pivot]
  # find the row with smallest ratio, 
  # so that after operation vector b can stay positive
  row = np.argmin(ratios)
  # transfrom row index to the row index of matrix A
  row = candrows[row]
  col = np.nonzero(A[row, card_basis])[0][0]
  new_pivot = card_basis[col]

  # In principle, could return new_pivot now can perform the matrix and vector
  # in parallel with ordinal pivot step (parallel is hard to realize on Windows)
  card_basis[col] = pivot

  b[row] /= A[row, pivot]
  A[row, :] /= A[row, pivot]

  # numba prange based implementation seems to be fast for small matrices.
  _update_A_b(A, b, row, pivot, np.nonzero(A[:, pivot])[0])

  # blas based one seems to be faster for large matrix.
  # _update_A_b_blas(A, b, row, pivot)

  # TODO: compare two update functions on a more powerful computer.

  return new_pivot


def ordinal_pivot(U, ord_basis, pivot, rowmin_locations, U_dom):
  """Ordinal pivot step.

  Remove `pivot` from an ordinal basis `basis`, and replace it with another
  column so that `ordinal_basis` is still an ordinal basis.

  Args:
    basis: an increasing array of indices.
    rowmin_locations: array of row index of rowmin location for each column in 
      the basis.
    U_dom: (m, n) boolean array of entry domination.

  Returns:
    The column that was introduced into basis
  """
  j = np.searchsorted(ord_basis, pivot)
  ru = rowmin_locations.pop(j)
  pivot_out = ord_basis.pop(j) # remove `pivot` from `basis`
  col_w_2_rowmin = my_argmin(U[ru, ord_basis])
  # Now column ord_basis[cn] has two row min.
  ri = rowmin_locations[col_w_2_rowmin]  # Our row of interest. 
  # We will use this row to find a new column to ordinal basis
  # row `ru` has a new row min at column ord_basis[cn]
  rowmin_locations[col_w_2_rowmin] = ru  # The row min location is updated
  _update_U_dom_row(U_dom, U, ru, ord_basis[col_w_2_rowmin])

  # Find columns that are ONLY dominated via row `ri`
  candcols = np.nonzero(~(np.any(U_dom[:ri, :], axis=0) | 
                          np.any(U_dom[ri+1:, :], axis=0)))[0]

  # Among these rows, choose the one with the largest utility
  new_col = np.argmax(U[ri, candcols]) # ri, candcols
  new_pivot = candcols[new_col]

  # Could return here for multithread implementation

  # Update basis and book-keeping variables
  j = np.searchsorted(ord_basis, new_pivot)
  ord_basis.insert(j, new_pivot)

  # update U_dom matrix.
  # row `ri` has a new row min at column new_pivot
  rowmin_locations.insert(j, ri)
  _update_U_dom_row(U_dom, U, ri, new_pivot)

  return new_pivot


def _scarf_pivot(A, U, b, verbose=False):
  """Scarf Algorithm Implementation.

  Find a "fractionally stable" extreme point of the polytope {x >= 0 : A x = b}.
  Iteratively apply ordinal and cardinal pivot to obtain a basis that is both
  cardinal and ordinal.

  Args:
    A: Constraint matrix of size (m, n). Must be numpy array with non-negative
       entries.
    U: Utility matrix of size (m, n). Must have numpy array with non-positive
       entries. 
    b: RHS vector of size (n,). Must be numpy array with non-negative entries.

  Returns:
    x: a real-valued vector of length (n,)
    basis: a feasible and stable basis, which is list of column indices of
       length (m,).
  """
  num_rows = len(b)
  A = A.astype(np.float64, order='C')  # for blas functions to work well
  b = b.astype(np.float64)
  # perturbation for cardinal pivot assumption to hold
  # exp(a) is a trancendental number if a is rational!
  # Ain't no polynomial got this solution.
  b += 1e-6 * np.exp(np.linspace(start=0.0, stop=1.0,
                                 num=num_rows, dtype=np.float64))
  # initialize cardinal and ordinal basis, as well as book keeping variables
  card_basis = list(range(num_rows))
  pivot_0 = np.argmax(U[0, num_rows:]) + num_rows
  ord_basis = list(range(1, num_rows)) + [pivot_0]

  # initialize the location (row) of rowmin for each column in the ordinal basis
  rowmin_locations = list(range(1, num_rows)) + [0]
  # initialize the domination matrix
  U_dom = np.concatenate(
    (np.diag([True] * num_rows), 
     np.tile(np.array([[True] + [False] * (num_rows - 1)]).T, 
             A.shape[1] - num_rows)),
    axis=1
  )

  # card_basis
  pivot_steps = 0 
  total_card_time, total_ord_time = 0.0, 0.0
  while True:
    if pivot_steps % 200 == 0 and verbose:
      print("current pivot step: #{0}".format(pivot_steps))
    start_time = time.time()
    pivot_1 = cardinal_pivot(A, b, card_basis, pivot_0)
    end_time = time.time()
    duration = end_time - start_time
    total_card_time += duration
    if pivot_1 == 0:
      break
    start_time = time.time()
    pivot_0 = ordinal_pivot(U, ord_basis, pivot_1, rowmin_locations, U_dom)
    end_time = time.time()
    duration = end_time - start_time
    total_ord_time += duration
    if pivot_0 == 0:
      break
    pivot_steps += 1
  
  return card_basis, pivot_steps, total_card_time, total_ord_time


def scarf_solve(A, U, b, verbose=False):
  """Scarf Algorithm Implementation.

  Find a "fractionally stable" extreme point of the polytope {x >= 0 : A x = b}.
  Iteratively apply ordinal and cardinal pivot to obtain a basis that is both
  cardinal and ordinal.

  Args:
    A: Constraint matrix of size (m, n). Must be numpy array with non-negative
       entries.
    U: Utility matrix of size (m, n). Must have numpy array with non-positive
       entries. 
    b: RHS vector of size (n,). Must be numpy array with non-negative entries.

  Returns:
    x: a real-valued vector of length (n,)
    basis: a list of column indices of length (m,)
  """
  if not A.shape == U.shape:
    raise ValueError("Matrix Dimention mismatch.")
  if not A.shape[0] == len(b):
    raise ValueError("Matrix Vector Dimention mismatch.")
  
  start_time = time.time()
  basis, pivot_steps, total_card_time, total_ord_time = _scarf_pivot(
      A, U, b, verbose)  # A, b are kept same after call.
  end_time = time.time()
  if verbose:
    print("Pivoted for {0} steps in {1:.3f}s.".format(
        pivot_steps, end_time - start_time))
    print("Avg cardinal Pivot: {0:.3f}ms.".format(
        1e3 * total_card_time / pivot_steps))
    print("Avg Ordinal Pivot: {0:.3f}ms.".format(
        1e3 * total_ord_time / pivot_steps))

  try:
    alloc = np.linalg.solve(A[:, basis], b)
  except np.linalg.LinAlgError:
    raise RuntimeError("Scarf algorithm cardinal pivoting failed due to "
                       "floating point precision problem. "
                       "(Yep this is very unlucky...)")
  return alloc, basis, pivot_steps


def iterative_rounding(basis, alloc, A, num_doctor_row, b, tol=1e-6, 
                       verbose=False):
  """Round a feasible and stable fractional solution to a new solution.

  Invented by T. Nguyen and R. Vohra, the Iterative Rounding Method performs a
    rounding procedure on a feasible and stable fractional solution of a stable
    matching problem and returns a integral stable matching which violates the 
    capacity constraint, but the violations are small. This implementation can 
    guarantee that no individual hospital would be overallocated by more than 2
    seats, and the total overallocation will be no more than 4 seats.

  Args:
    basis: The feasible and dominating basis (with no slack variable) to start with.
    alloc: The allocation vector with only the elements in basis.
    A: Doctor constraint matrix, Hospital capacity constraint matrix.
    num_doctor_row: A[:num_doctor_row, :] is the doctor constraint matrix.
    hospital_cap: Hospital capacity vector.
  """

  # basis_np = np.array(basis)
  # non_slack = basis_np >= A.shape[0]

  # A = A[:, basis_np[non_slack]].toarray()
  A = A[:, basis].toarray()
  aggh = np.sum(A[num_doctor_row:, :], axis=0, keepdims=True)
  aggb = np.sum(b[num_doctor_row:])

  A = np.concatenate((A, aggh), axis=0)
  b = np.concatenate((b, [aggb]))
  x = np.round(alloc, 4 - np.log10(tol).astype(np.int64))
 
  active_rows = list(range(num_doctor_row, A.shape[0]))
  agg_still_alive = True
  while True:
    x_is_int = is_int_vec(x, 1e-4 * tol)
    # print("\nIter! x = ", np.round(x[~x_is_int], 10))
    # print("Iter! id = ", np.nonzero(~x_is_int)[0])
    # print("Active rows: ", np.take(A[:, ~x_is_int], active_rows, axis=0))
    b1 = b - rd(np.dot(A[:, x_is_int], x[x_is_int]))
    if np.all(x_is_int):
      # return basis_np[non_slack], rd(x)
      break
    binding = np.dot(A, x) > b1 - tol 
    x_is_frac = ~x_is_int
    candrows = binding[active_rows] & (
        np.dot(A[active_rows, :], x_is_frac) <= 3)
    # print("candrows: ", [active_rows[j] for j in np.nonzero(candrows)[0]])
    if np.any(candrows):
      elim = np.nonzero(candrows)[0][0]
      active_rows.pop(elim)
      # print("# Elim!")
    elif agg_still_alive:
      contain_frac = ~binding[:num_doctor_row] & (
          np.dot(A[:num_doctor_row, ~slack], x_is_frac[~slack]) > 0)
      if np.sum(contain_frac) <= 2:
        active_rows[-1:] = []
        agg_still_alive = False
    else:
      raise RuntimeError("Something Wrong...")

    binding_doctors = np.nonzero(binding[:num_doctor_row])[0]
    userows = np.concatenate(
        (np.nonzero(~binding[:num_doctor_row])[0], active_rows))
    A_eq, b_eq = A[binding_doctors, :], b1[binding_doctors]
    A_ub, b_ub = A[userows, :], b1[userows]
    A_eq, b_eq = A_eq[b_eq > 0, :], b_eq[b_eq > 0]
    A_ub, b_ub = A_ub[b_ub > 0, :], b_ub[b_ub > 0]

    # print("The program to solve: c = ", aggh[0, x_is_frac])
    # print("A_ub = ", A_ub[:, x_is_frac])
    # print("A_eq = ", A_eq[:, x_is_frac])
    # print("b_ub = ", b_ub)
    # print("b_eq = ", b_ub)

    res = spopt.linprog(
      c= -aggh[0, x_is_frac], A_ub=A_ub[:, x_is_frac], b_ub=b_ub,
      A_eq=A_eq[:, x_is_frac], b_eq=b_eq, method="simplex")
    if not res.success:
      raise RuntimeError("Linear program failed!")
    x[x_is_frac] = res.x 
    # print("solution: ", res.x)

  # Intergral solution has been found!
  return rd(x)


def rd(vec):
  return np.round(vec).astype(np.int32)


def is_int(vec, tol=1e-10):
  return np.all(is_int_vec(vec, tol))


def is_int_vec(vec, tol=1e-10):
  return np.abs(vec - np.round(vec)) <= tol



"""Scarf algorithm library for python3.

Create and solve a stable matching instance with couple using Scarf's algorithm.

Dengwang Tang <dwtang@umich.edu>
"""

import numpy as np
from scipy import sparse as sp

import scarf.core

__all__ = [
    "ScarfInstance", "create_instance", "solve", "round"
]


def _is_unique(li):
  """Check if elements of a list is unique.""" 
  return len(li) == len(set(li))


def _sanity_check(num_hospital,
                  single_pref_list, couple_pref_list, hospital_pref_list):
  """Sanity check for construction of instances.

  Raises:
    ValueError if any of the preference list has duplicate / has improper index.
  """
  num_single = len(single_pref_list)
  num_couple = len(couple_pref_list)
  for li in couple_pref_list:
    if not _is_unique(li):
      raise ValueError("Duplicate hospital pair in couple's preference list.")
    for h0, h1 in li:
      if h0 >= num_hospital:
        raise ValueError("Hospital index out of range in couple's "
                         "preference list.")
      if h1 >= num_hospital:
        raise ValueError("Hospital index out of range in couple's "
                         "preference list.")
      if h0 < 0 and h1 < 0:
        raise ValueError("Hospital pair not admissible.")
  
  for li in single_pref_list:
    if not _is_unique(li):
      raise ValueError("Duplicate hospital in single's preference list.")
    for h in li:
      if not 0 <= h < num_hospital:
        raise ValueError("Hospital index out of range in single's "
                         "preference list.")

  if not isinstance(hospital_pref_list[0], list):
    hospital_pref_list = [hospital_pref_list]
  for li in hospital_pref_list:
    if not _is_unique(li):
      raise ValueError("Duplicate doctor in hospital's preference list.")
    if not len(li) == num_single + 2 * num_couple:
      raise ValueError("Hospital must rank all doctors.")
    for i in li:
      if isinstance(i, int):
        if not 0 <= i < num_single:
          raise ValueError("Single index out of range in hospital's "
                           "preference list.")
      elif isinstance(i, tuple):
        if not 0 <= i[0] < num_couple:
          raise ValueError("Couple index out of range in hospital's "
                           "preference list.")
        if not 0 <= i[1] <= 1:
          raise ValueError("The second element of a tuple representing a member"
                           " of a couple must be either 0 or 1.")


def _tuple_2_id(num_single, idx):
  """Transfrom from tuple representation to id.

  If idx is a tuple representation of a member of a couple, transform it to 
  id representation. Otherwise do nothing.

  Raises:
    TypeError if idx is neither an integer or a tuple
  """
  if isinstance(idx, int):
    return idx
  elif isinstance(idx, tuple):
    return num_single + 2 * idx[0] + idx[1]
  else:
    raise TypeError("Cannot recognize idx as single or couple")


def _pref_list_2_score_list(num_single, num_couple, num_hospital, hospital_pref_list):
  """Transform one hospital preferences list to a list of scores.

  All returned scores are in range (-num_applicant, 0] * num_hosp_pairs - 1, 
  this allows comparison with 0 in sparse matrices, as well as leaving space
  for tie breaking by perturbing the scores with couple's will. 

  Args:
    num_single: number of single doctors
    num_couple: number of couples
    num_hospital: number of hospitals
    hospital_pref_list: a list of preference lists, where each preference
    list is a list of ints (singles) and tuples (couple, member_id)

  Returns:
    1. Score of singles
    2. Score of first members of couples
    3. Score of second member of couples
  """
  is_identical_pref = not isinstance(hospital_pref_list[0], list)
  num_hospital_pair = (num_hospital + 1) ** 2 - 1
  doctor_scores = np.zeros(
    (num_hospital + 1, num_single + 2 * num_couple), dtype=np.int32)
  for h in range(num_hospital):
    # transform from tuple to idx

    li = list(map(
      lambda x: _tuple_2_id(num_single, x), hospital_pref_list[h]
    ))
    doctor_scores[h][li] = - np.arange(len(li))

  doctor_scores = doctor_scores * num_hospital_pair - 1
  return (doctor_scores[:num_hospital, :num_single],
          doctor_scores[:, np.arange(
              num_single, num_single + 2 * num_couple, 2)],
          doctor_scores[:, np.arange(
              num_single + 1, num_single + 2 * num_couple, 2)]) 


class HospPrefList():
  def __init__(self, hospital_pref_list):
    self.hospital_pref_list = hospital_pref_list
    self.is_ihp = not isinstance(hospital_pref_list[0], list)
    if self.is_ihp:
      self.doctor_lookup = {hospital_pref_list[i]: i+1
                            for i in range(len(hospital_pref_list))}
    else:
      self.doctor_lookup = [{hospital_pref_list[h][i]: i+1
                            for i in range(len(hospital_pref_list[h]))}
                            for h in range(len(hospital_pref_list))]   

  def __getitem__(self, h):
    if self.is_ihp:
      return self.hospital_pref_list
    else:
      return self.hospital_pref_list[h]

  def rank(self, h, s_or_c):
    if self.is_ihp:
      return self.doctor_lookup[s_or_c]
    else:
      return self.doctor_lookup[h][s_or_c]



class ScarfInstance():
  """Stable matching problem instance.

  An object storing all the preferences of doctors and hospitals, as well as the
  capacity of hospitals.

  Attributes:
    num_single: Number of single doctors in the problem.
    num_couple: Number of couples in the problem. 
    num_applicants: Number of total doctors, equals `num_single + 2 * num_couple`.
    single_pref_list: A list of preference list of singles.
      `single_pref_list[i][j]` is the j-th most prefered hospital of the i-th
      single doctor. Any hospital not in `single_pref_list[i]` is considered as
      worse than unemployment option by the i-th single doctor.
    couple_pref_list: A list of preference list of couples.
      `couple_pref_list[i][j]` is the j-th most prefered hospital pair of the
      i-th couple. Any hospital pair not in `couple_pref_list[i]` is considered as
      worse than the 'both unemployed' option by the i-th couple.
    hospital_pref_list: A `HospPrefList` object.
      `hospital_pref_list[i][j]` is the j-th most prefered individual doctor of
      the i-th hospital. Single doctors are represented by an integer, while the
      first member of the i-th couple is represented by a tuple (i, 0), and the 
      second member is represented by (i, 1).
    hospital_cap: A list of capacities of hospitals.
    pair_list: list of all admisible allocation plans. (i.e. no single/couple/hospital
      involved in this plan rate unemployment/'both unemployed'/'empty seat'
      higher than this plan.) A plan is a tuple `(s, h0)` or `(c, h0, h1)` where
      `s` is in [-1, num_singles), `c` is in [0, num_couples), `hj` is in
      [-1, num_hospitals), where -1 indicates the outside option.
    A: The constraint/assignment matrix of `num_single + num_couple + num_hospital`
      rows and `len(pair_list)` columns. Rows represent single doctors, couples,
      and hospitals, in this order. Columns represent the allocation plans.
    U: The utility matrix with the same size as A. The rows and columns represents
      the same as above.
  """
  def __init__(self, single_pref_list,
               couple_pref_list, hospital_pref_list, hospital_cap):
    """
    Create an instance of stable matching problem.

    Construct a stable matching problem using preference lists and hospital 
    capacity vector.

    Args:
      single_pref_list: list of list of hospital indices. 
      couple_pref_list: list of list of tuple of two hospital indices. Use -1
        to indicate the option of unemployment.
      hospital_pref_list: 
        - either a list of list of integers and tuples, where integers
          represent the index of a single doctor, and a tuple (c, j) represent
          the j-th member of couple c (j is either 0 or 1)
        - or a list of integers and tuples for identical hospital preference
      hospital_cap: a list indicating the number of seats in each hospital.
    """
    
    self.num_single = len(single_pref_list)
    self.num_couple = len(couple_pref_list)
    self.num_hospital = len(hospital_cap)

    self.num_applicant = self.num_single + 2 * self.num_couple
    self._nhp = (self.num_hospital + 1) ** 2 - 1

    # store the data
    self.single_pref_list = single_pref_list
    self.couple_pref_list = couple_pref_list
    self.hospital_pref_list = HospPrefList(hospital_pref_list)
    self.hospital_cap = hospital_cap

    # creates mapping from matrix index to allocation plan
    self._setup_pair_list()
    self.A, self.U = self._create_matrices()

  def __repr__(self):
    s = "<ScarfInstance with {s} singles, {c} couples, and {h} hospitals>".format(
        s=self.num_single, c=self.num_couple, h=self.num_hospital
    )
    return s

  def _setup_pair_list(self):
    """Creates the list of pairs.

    Creates a list of doctor(s)-hospital(s) pairs (i.e. (s, h), (c, h0, h1))
    which indicates represents the column of constraint and utility matrix.
    """
    slack_list = [(s, -1) for s in range(self.num_single)]
    slack_list += [(c, -1, -1) 
                   for c in range(self.num_couple)]
    slack_list += [(-1, h) for h in range(self.num_hospital)]
    single_pair_list = []
    for s in range(self.num_single):
      single_pair_list += [(s, h) for h in self.single_pref_list[s]]
    couple_pair_list = []
    for c in range(self.num_couple):
      couple_pair_list += [(c, h0, h1) for h0, h1 in self.couple_pref_list[c]]
    self.pair_list = slack_list + single_pair_list + couple_pair_list
    self._lookup = {self.pair_list[j]: j for j in range(len(self.pair_list))}
    self._nsl, self._nps = len(slack_list), len(single_pair_list), 
    self._npc = len(couple_pair_list)

  def _slack_list(self):
    """Returns a list of allocation plans corresponding to outside options."""
    return self.pair_list[:self._nsl]

  def admissible_plans(self):
    """Returns admissible allocation plans that are not outside options.

    Returns admissible allocation plans that involves at least one doctor and one
    hospital.
    """
    return self.pair_list[self._nsl:]

  def _single_pair_list(self, j=-1):
    """Returns admisible allocation plans involving single doctors."""
    if j < 0:
      return self.pair_list[self._nsl: self._nsl + self._nps]
    else:
      return self.pair_list[self._nsl + j]

  def _couple_pair_list(self, j=-1):
    """Returns admisible allocation plans involving couples."""
    if j < 0:
      return self.pair_list[self._nsl + self._nps:]
    else:
      return self.pair_list[self._nsl + self._nps + j]

  def _create_single_ps_matrices(self):
    # Construct single - (single, hospital) utility matrix.
    I = [s for s, h in self._single_pair_list()]
    J = range(self._nps)
    # For each single, assign score from high to low.
    W = np.ones(self._nps, dtype=np.int8)  # for A
    if self.num_single > 0:
      V = np.concatenate([-np.arange(len(self.single_pref_list[r])) - 1
                          for r in range(self.num_single)])  # for U
    else:
      V = np.arange(0)
    A_s_ps = sp.coo_matrix(
        (W, (I, J)),  shape=(self.num_single, self._nps), dtype=np.int8)
    U_s_ps = sp.coo_matrix(
        (V, (I, J)), shape=(self.num_single, self._nps), dtype=np.int32)
    return A_s_ps, U_s_ps

  def _create_couple_pc_matrices(self):
    """Create couple - (couple, hospitals) utility matrix.

    Returns:
      1. The matrix
      2. A 1-D numpy array of scores couples assigned to pairs, for tie breaking.
    """
    I = [c for c, h0, h1 in self._couple_pair_list()]
    J = range(self._npc)
    W = np.ones(self._npc, dtype=np.int8)
    # For each couple, assign score from high to low.
    if self.num_couple > 0:
      V = np.concatenate([-np.arange(len(self.couple_pref_list[r])) - 1
                          for r in range(self.num_couple)])
    else:
      V = np.arange(0)
    A_c_pc = sp.coo_matrix(
        (W, (I, J)), shape=(self.num_couple, self._npc), dtype=np.int8)
    U_c_pc = sp.coo_matrix(
        (V, (I, J)), shape=(self.num_couple, self._npc), dtype=np.int32)
    return A_c_pc, U_c_pc, V + 1

  def _create_hospital_ps_matrices(self, single_scores):
    I = [h for s, h in self._single_pair_list()]
    J = range(self._nps)
    W = np.ones(self._nps, dtype=np.int8)
    # The utility value is the score hospital assigned to single.
    V = [single_scores[h][s] for s, h in self._single_pair_list()]
    # Construct the matrix
    A_h_ps = sp.coo_matrix(
        (W, (I, J)), shape=(self.num_hospital, self._nps), dtype=np.int8)
    U_h_ps = sp.coo_matrix(
        (V, (I, J)), shape=(self.num_hospital, self._nps), dtype=np.int32)
    return A_h_ps, U_h_ps

  def _create_hospital_k_pc_matrices(self, couple_scores, V_c_pc, k):
    Ik = [p[k+1] for p in self._couple_pair_list() if p[k+1] >= 0]
    Jk = [j for j in range(self._npc)
          if self._couple_pair_list(j)[k+1] >= 0]
    Wk = [1 for j in range(self._npc)
          if self._couple_pair_list(j)[k+1] >= 0]
    # The utility value is the score hospital assigned to member 0 of couple.
    Vk = np.array([couple_scores[p[k+1]][p[0]]
                   for p in self._couple_pair_list()
                   if p[k+1] >= 0]) + V_c_pc[Jk]
          # adding couple's utility for tie breaking.
    A_hk_pc = sp.coo_matrix(
        (Wk, (Ik, Jk)), shape=(self.num_hospital, self._npc), dtype=np.int8)
    U_hk_pc = sp.coo_matrix(
        (Vk, (Ik, Jk)), shape=(self.num_hospital, self._npc), dtype=np.int32)
    return A_hk_pc, U_hk_pc

  def _create_matrices(self):
    """Create constraint and utility matrices for scarf pivoting."""
    
    A_s_ps, U_s_ps = self._create_single_ps_matrices()
    A_s_pc = U_s_pc = sp.coo_matrix(
        (self.num_single, self._npc), dtype=np.int8)
    A_c_ps = U_c_ps = sp.coo_matrix(
        (self.num_couple, self._nps), dtype=np.int8)
    A_c_pc, U_c_pc, V_c_pc = self._create_couple_pc_matrices()

    single_scores, couple_0_scores, couple_1_scores = _pref_list_2_score_list(
        self.num_single, self.num_couple, self.num_hospital,
        self.hospital_pref_list)
    A_h_ps, U_h_ps = self._create_hospital_ps_matrices(single_scores)
    A_h0_pc, U_h0_pc = self._create_hospital_k_pc_matrices(
        couple_0_scores, V_c_pc, 0)
    A_h1_pc, U_h1_pc = self._create_hospital_k_pc_matrices(
        couple_1_scores, V_c_pc, 1)

    A_h_pc = A_h0_pc + A_h1_pc
    U_h_pc = U_h0_pc.minimum(U_h1_pc)  # Take the worse of the two members

    AL = sp.eye(self._nsl, dtype=np.int8)
    # Create slack variable utility
    min_u = - self.num_applicant * self._nhp  # the minimum U can achieve
    UL = sp.diags(
        [-self.num_hospital - 1 for _ in range(self.num_single)] + 
        [-self._nhp - 1 for _ in range(self.num_couple)] + 
        [min_u - 1 for _ in range(self.num_hospital)], dtype=np.int32
    )

    A = sp.hstack([AL, sp.vstack(
        [
            sp.hstack([A_s_ps, A_s_pc]),
            sp.hstack([A_c_ps, A_c_pc]),
            sp.hstack([A_h_ps, A_h_pc])
        ])
    ], format="csr")
    U = sp.hstack([UL, sp.vstack(
        [
            sp.hstack([U_s_ps, U_s_pc]),
            sp.hstack([U_c_ps, U_c_pc]),
            sp.hstack([U_h_ps, U_h_pc])
        ])
    ], format="csr")
    return A, U

  def full_A(self):
    """Obtains constraint matrix as a full matrix."""
    return self.A.toarray()

  def full_U(self):
    """Obtains utility matrix as a full matrix."""
    return self.U.toarray()

  def full_b(self):
    """Obtains right-hand-side vector as a 1-d array."""
    return np.concatenate(
        (np.ones(self.num_single + self.num_couple, dtype=np.int32),
         np.array(self.hospital_cap, dtype=np.int32)))

  def hospital_rank_by_single(self, s, h):
    """Obtains the ranking of hospital h by single s.

    The most prefered hospital is of rank 1, the second is 2, and so on...

    Args:
      s: Index of single doctor.
      h: Index of hospital.

    Returns:
      Ranking of hospital.
    """
    j = self._lookup.get((s, h))
    if j is None:
      return self.num_hospital
    else:
      return -self.U[s, j]

  def hospital_rank_by_couple(self, c, hs):
    """Obtains the ranking of hospital pair (h0, h1) by single s.

    The most prefered hospital pair is of rank 1, the second is 2, and so on...

    Usage: `S.hospital_rank_by_couple(c, (h0, h1))`
    Args:
      c: Index of couple.
      h0: Index of hospital, corresponds to the first member of the couple.
      h1: Index of hospital, corresponds to the second member of the couple.

    Returns:
      Ranking of hospital pair.
    """
    h0, h1 = hs
    j = self._lookup.get((c, h0, h1))
    if j is None:
      return self._nhp + 1
    else:
      return -self.U[self.num_single + c, j]

  def doctor_rank_by_hospital(self, h, s_or_c):
    """Obtains the ranking of individual doctor by a hospital.

    The most prefered doctor is of rank 1, the second is 2, and so on...
    
    Args:
      h: Index of hospital.
      s_or_c: either an integer representing the index of a single doctor, or
        a tuple `(c, j)` j=0,1 indicating a member of a couple.

    Returns:
      The ranking of doctor.
    """
    return self.hospital_pref_list.rank(h, s_or_c)


def _s_select(pair_list, s):
  return [p for p in pair_list if len(p) == 2 and p[0] == s and p[1] >= 0]


def _is_c_related(p, c):
  return len(p) == 3 and p[0] == c and (p[1] >= 0 or p[2] >= 0)


def _c_select(pair_list, c):
  return [p for p in pair_list if _is_c_related(p, c)]


def _is_h_related(p, h):
  return (p[1] == h or p[-1] == h) and p[0] >= 0


def _h_select(pair_list, h):
  return [p for p in pair_list if _is_h_related(p, h)]


def _get_member(p, h):
  """Return the doctors allocated to hospital h via a pair p.
  """
  if len(p) == 2: # represents allocation to a single doctor
    return [p[0]]
  # otherwise, its an allocation for a couple
  if p[1] == h and p[2] != h:
    return [(p[0], 0)]
  if p[1] != h and p[2] == h:
    return [(p[0], 1)]
  if p[1] == h and p[2] == h:
    return [(p[0], 0), (p[0], 1)]


class ScarfSolution():
  """Solution of a stable matching instance.

  A possibly fractional matching of doctors to hospitals, which may or may not
  violate the capacity constraint.

  One can directly access this object to obtain the solution:
  e.g. for a `ScarfSolution` sol,
    sol[(s, h)] is the fractional indicator (weight) of the allocation plan
      'single s to hospital h'.
    `sol["s20"]` or `sol.s(20)` is a list of hospitals allocated to the 20-th
    (0-indexed) single doctor with non-zero weight/fraction. If the solution is integral,
    then this list has only one element, which is the hospital allocated to the
    20-th single doctor.
    `sol["c12"]` or `sol.c(12)` is a list of hospital pairs allocated to the 12-th
    (0-indexed) couple with non-zero weight/fraction. If the solution is integral,
    then this list has only one element, which is the hospital pair allocated to the
    20-th single doctor.
    `sol["h0"]` or `sol.h(0)` is a list of doctors (single doctor or members of couples)
    allocated to the 0-th hospital with non-zero weight/fraction. If the solution is integral,
    then it is the list of doctors allocated to the 0-th hospital.

  Attributes:
    num_pivots: number of pivots performed.
    is_int: True if the solution is integral.
    act_hospital_cap: array of number of allocated seats in each hospitals after
      allocation.

  Other Attributes for the interest of researchers: (where the associated `ScarfInstance` object
    `ins` is needed to understand what these attributes represent.)
    basis: list of column indices of the matrix ins.A, indicating the support of
      the matching vector.
    alloc: allocation vector, where `alloc[i]` is the fractional indicator of the
      doctor-hospital pair `ins.pair_list[self.basis[i]]`.
  """
  def __init__(self, S, basis, alloc, num_pivots=-1, tol=1e-6):
    """
    Args:
      S: a `ScarfInstance` object.
      basis: list of column indices of the matrix S.A, indicating the support of
        the matching vector.
      alloc: allocation vector, where `alloc[i]` is the fractional indicator of the
        doctor-hospital pair `ins.pair_list[self.basis[i]]`.
      num_pivots: number of pivots in the algorithm.
      tol: tolerance for integral testing.
    """
    basis_np = np.array(basis, dtype=np.int64)
    self.basis = basis_np[basis_np >= S._nsl]
    self.is_int = scarf.core.is_int(alloc, tol)
    alloc_np = np.array(alloc, dtype=np.int64 if self.is_int else np.float64)
    self.alloc = alloc_np[basis_np >= S._nsl]

    self.act_hospital_cap = np.dot(
        np.take(S.A[S.num_single + S.num_couple:, :].toarray(),
                self.basis, axis=1), self.alloc)
    self.num_pivots = num_pivots
    # For solution lookup
    self._num_hospital = S.num_hospital
    self._solution_map = {S.pair_list[self.basis[i]]: np.round(self.alloc[i], 6)
                          for i in range(len(self.basis)) if self.alloc[i] > tol}
    _pair_list = self._solution_map.keys()
    self.single_allocations = [
        _s_select(_pair_list, s) for s in range(S.num_single)] 
    self.couple_allocations = [
        _c_select(_pair_list, c) for c in range(S.num_couple)]
    self.hospital_allocations = [
        _h_select(_pair_list, h) for h in range(S.num_hospital)]

  def __repr__(self):
    strlist = ["<ScarfSolution of {s} singles, {c} couples, {h} hospitals".format(
        s=len(self.single_allocations), c=len(self.couple_allocations),
        h=self._num_hospital
    )] 
    strlist += ["\twith {0} solution>".format(
        "integral" if self.is_int else "NON integral")
    ]
    return "\n".join(strlist)

  def __str__(self):
    strlist = ["Solution for instance with {s} singles, {c} couples, {h} hospitals.".format(
        s=len(self.single_allocations), c=len(self.couple_allocations),
        h=self._num_hospital
    )]
    strlist += ["Status: {0} solution.".format(
        "integral" if self.is_int else "NON integral")
    ]
    if self.is_int:
      for s in range(len(self.single_allocations)):
        strlist += ["Doctor {0} is assigned to hospital {1}".format(s, self.s(s)[0])]
      for c in range(len(self.couple_allocations)):
        h0, h1 = self.c(c)[0]
        strlist += ["First member of Couple {0} is assigned to hospital {1}".format(c, h0)]
        strlist += ["Second member of Couple {0} is assigned to hospital {1}".format(c, h1)]
      for h in range(self._num_hospital):
        strlist += ["Hospital {0} is assigned with the following doctors: {1}".format(h,
          ", ".join(str(x) for x in self.h(h)))]
    return "\n".join(strlist)


  def __getitem__(self, s):
    """Get matched hospital/doctor.

    Args:
      s: either a tuple indicating a doctor(s)-hospital(s) pair, or a string
         indicating a single (e.g. "s10") or a couple ("c2") or a hospital
         (e.g. "h12")

    Returns:
      a number between 0 and 1 (allocation fraction) if `s` is a doctor(s)-hospital(s)
         pair; a dictionary from hospital/hospitals/doctor to allocation fraction
         if `s` is a string indicating single/couple/hospital.

    Raises:
      ValueError if input string is not parsable.
      TypeError if input is neither a tuple nor a string.
    """
    if isinstance(s, tuple):
      if s in self._solution_map:
        return self._solution_map[s]
      else:
        return 0.0
    elif isinstance(s, str):
      if s.startswith("s"):
        return self.get_single_allocation(int(s[1:]))
      elif s.startswith("c"):
        return self.get_couple_allocation(int(s[1:]))
      elif s.startswith("h"):
        return self.get_hospital_allocation(int(s[1:]))
      else:
        raise ValueError("Cannot parse input.")
    else:
      raise TypeError("Unrecognized index type.")

  def get_single_allocation(self, s, show_fractions=False):
    """Get the hospital (or hospitals and/or their corresponding fractions) matched to
       single doctor s

       Returns:
         A list of hospital indices if show_fractions=False
         A dictionary mapping hospital indices to their corresponding fractions if show_fractions=True
       
    """
    if show_fractions:
      return {p[1]: self._solution_map[p] for p in self.single_allocations[s]}
    else:
      return [p[1] for p in self.single_allocations[s]]

  s = get_single_allocation

  def get_couple_allocation(self, c, show_fractions=False):
    """Get the pairs of hospital (or pairs of hospitals and/or their corresponding fractions) matched to
       couple c

       Returns:
         A list of hospital indices pairs if show_fractions=False
         A dictionary mapping hospital indices pairs to their corresponding fractions if show_fractions=True
       
    """
    if show_fractions:
      return {p[1:]: self._solution_map[p] for p in self.couple_allocations[c]}
    else:
      return [p[1:] for p in self.couple_allocations[c]]

  c = get_couple_allocation

  def get_hospital_allocation(self, h, show_fractions=False):
    """Get the doctors (or doctors and/or their corresponding fractions) matched to hospital h

       Returns:
         A list of hospital indices pairs if show_fractions=False
         A dictionary mapping hospital indices pairs to their corresponding fractions if show_fractions=True
       
    """

    if show_fractions:
      res = {}
      for p in self.hospital_allocations[h]:
        for member_of_c in _get_member(p, h):
          res[member_of_c] = self._solution_map[p]
      return res
    else:
      res = []
      for p in self.hospital_allocations[h]:
        res += _get_member(p, h)
      return res

  h = get_hospital_allocation


def create_instance(single_pref_list, couple_pref_list,
                    hospital_pref_list, hospital_cap):
  """Create an stable matching problem instance.

    Args:
      single_pref_list: list of list of hospital indices. 
      couple_pref_list: list of list of tuple of two hospital indices. Use -1
        to indicate the option of unemployment.
      hospital_pref_list: 
        - either a list of list of integers and tuples, where integers
          represent the index of a single doctor, and a tuple (c, j) represent
          the j-th member of couple c (j is either 0 or 1)
        - or a list of integers and tuples for identical hospital preference
      hospital_cap: a list indicating the number of seats in each hospital.  
  """
  _sanity_check(num_hospital=len(hospital_cap),
                single_pref_list=single_pref_list,
                couple_pref_list=couple_pref_list,
                hospital_pref_list=hospital_pref_list)
  return ScarfInstance(single_pref_list, couple_pref_list,
                       hospital_pref_list, hospital_cap)


def solve(ins, verbose=False):
  """Solve a stable matching instance with Scarf's algorithm.

  Args:
    ins: a `ScarfInstance` object.
    verbose: bool, optional
      If set to True, extra information will be printed when running the
      algorithm. Default is False.

  Returns:
    sol: a `ScarfSolution` object.
  """
  alloc, basis, num_pivots = scarf.core.scarf_solve(
      A=ins.full_A(),
      U=ins.full_U(),
      b=ins.full_b(),
      verbose=verbose
  )
  sol = ScarfSolution(S=ins, alloc=alloc, basis=basis, num_pivots=num_pivots)
  if verbose:
    print("Solution {0} integral.".format("is" if sol.is_int else "is NOT"))
  return sol


def round(sol, ins):
  """Round a fractional stable matching to an integral stable matching.

  Perform Iterative Rounding (IR) algorithm [1] on a fractional stable matching
  to obtain an integral stable matching with respect to a new hosptial capacity
  vector close to the original capacity vector.

  [1] Nguyen, Thanh, and Rakesh Vohra. "Near-feasible stable matchings with couples."
  American Economic Review 108.11 (2018): 3154-69.

  Args:
    sol: A `ScarfSolution` object associated with `ScarfInstance` ins
    ins: A `ScarfInstance` object.

  Returns:
    int_sol: A `ScarfSolution` object which represents an integral solution.
  """
  int_alloc = scarf.core.iterative_rounding(
      alloc=sol.alloc, basis=sol.basis, A=ins.A,
      num_doctor_row=ins.num_single + ins.num_couple, b=ins.full_b())
  return ScarfSolution(S=ins, alloc=int_alloc, basis=sol.basis,
                          num_pivots=sol.num_pivots)